import pandas as pd
import parse

from filefisher.filters import priority_filter

# select preferred grid; order indicates priority
VALID_GRIDS = ("gn", "gr", "gr1", "gm")


def parse_ens(filelist):

    ens = filelist.df["ens"]

    # for cmip6
    if "f" in ens.iloc[0]:
        parser = parse.compile("r{r}i{i}p{p}f{f}")
    # for cmip5
    else:
        parser = parse.compile("r{r}i{i}p{p}")

    out = list()
    for one_ens in ens:
        parsed = parser.parse(one_ens)

        # TODO: assumes this always works
        out.append(list(parsed.named.values()))

    keys = list(parsed.named.keys())

    df = pd.DataFrame(out, columns=keys)

    for key in df.columns:
        filelist.df[key] = df[key].values

    return filelist


def create_ensnumber(filelist, keys=None):

    if keys is None:
        keys = ("exp", "table", "varn", "model")

    keys = list(keys)
    df = filelist.df
    multiindex = pd.MultiIndex.from_frame(df[keys])

    df["ensnumber"] = -1

    for idx in multiindex.unique():

        sel = multiindex == idx
        numbers = list(range(sel.sum()))
        df.loc[sel, "ensnumber"] = numbers

    filelist.df = df
    return filelist


def ensure_unique_grid(filelist):
    """ensure there is only one grid per simulation"""

    # each simulation must be unique in the combination of these keys
    keys = ("exp", "table", "varn", "model", "ens")

    return priority_filter(filelist, column="grid", order=VALID_GRIDS, groupby=keys)



import functools

import pandas as pd

from filefisher._utils import emit_user_level_warning


def _wrap_filecontainer(func):
    @functools.wraps(func)
    def _inner(*args, **kwargs):

        from filefisher import FileContainer

        obj, *args = args

        is_filecontainer = False
        if isinstance(obj, FileContainer):
            obj = obj.df
            is_filecontainer = True

        obj = func(obj, *args, **kwargs)

        if is_filecontainer:
            obj = FileContainer(obj)

        return obj

    return _inner


@_wrap_filecontainer
def priority_filter(obj, column, order, *, on_missing="raise", groupby=None):
    """filter a dataframe on for nonunique entries according to a priority list

    Parameters
    ----------
    obj : pd.DataFrame | FileContainer
        Pandas DataFrame or FileContainer to filter.
    column : str
        The columt to apply the priority filter to.
    order : list of str
        The priority order.
    on_missing : "raise" | "warn" | "ignore", default "raise"
        Behaviour if none of the elements is found.
    groupby : None | list of str, default None
        Which columns to groupby over for the priority filter. Per default it uses all
        columns except `column`.

    """

    if on_missing == "error":
        emit_user_level_warning(
            "on_missing value 'error' has been renamed to 'raise'", FutureWarning
        )
        on_missing = "raise"

    if on_missing not in ["raise", "warn", "ignore"]:
        raise ValueError(
            f"Unknown value for 'on_missing': '{on_missing}'. Must be one of 'raise', 'warn' or 'ignore'."
        )

    if column not in obj.columns:
        raise ValueError(f"column ('{column}') must be available in df")

    if groupby is None:
        groupby = set(obj.columns) - {column, "filename"}

    if column in groupby:
        raise ValueError(f"`groupby` may not contain column ('{column}')")

    groupby = list(groupby)

    # create a MultiIndex -> can directly index over multiple keys
    mi = pd.MultiIndex.from_frame(obj[groupby])

    # there are models with more than one grid
    if mi.has_duplicates:

        obj = _prioritize(obj, column, order, on_missing, mi)

        # double check
        mi = pd.MultiIndex.from_frame(obj[groupby])

        if mi.has_duplicates:
            raise ValueError("Something went wrong")

    return obj


def _prioritize(df, key, order, on_missing, multiindex):

    out = list()

    # loop entries
    for idx in multiindex.unique():

        one_ = df.iloc[multiindex.get_locs(idx)]

        # select the first that is found
        for element in order:
            if element in one_[key].values:

                selected = one_[one_[key] == element]

                if len(selected) != 1:
                    idx_string = one_.to_string()
                    raise ValueError(
                        f"Found more than one `df[{key}] == '{element}'` for\n{idx_string}"
                    )

                out.append(selected)
                break

        # no entry was found
        else:
            idx_string = one_.to_string()

            if on_missing == "raise":
                raise ValueError(
                    f"Did not find any element from the priority list for\n{idx_string}"
                )
            elif on_missing == "warn":
                emit_user_level_warning(
                    f"Did not find any element from the priority list for\n{idx_string}"
                )
            elif on_missing == "ignore":
                pass

    df = pd.concat(out, axis=0)

    return df



import itertools
import pathlib
import re
import sys
import warnings

import pandas as pd


def _find_keys(string):
    """find keys in a format string

    find all keys enclosed by curly brackets

    >>> _find_keys("/path/{var_name}/{year}") == ("var_name", "year")
    True

    >>> _find_keys("/path/{var_name}/{year:d}") == ("var_name", "year")
    True
    """

    # match group
    pattern = (
        r"\{"
        r"([A-Za-z0-9_]+)"  # capturing group with one or more characters, number or _
        r"(?::.*?)?"  # non-capturing group, non greedy matching any char, zero or once
        r"\}"
    )

    keys = re.findall(pattern, string)
    keys = tuple(pd.Series(keys).unique())

    return keys


def atoi(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):
    """key for natural sorting order

    Examples
    --------
    >>> l = ['a10', 'a1']
    >>> l.sort(key=natural_keys)
    >>> l
    ['a1', 'a10']

    References
    ----------
    http://nedbatchelder.com/blog/200712/human_sorting.html
    """
    return [atoi(c) for c in re.split(r"(\d+)", text)]


def product_dict(**kwargs):
    """generate list of dictionaries with all possible combinations

    Examples
    --------
    >>> list(product_dict(**{"a": [1, 2], "b": [3, 4], "c": [5]}))
    [{'a': 1, 'b': 3, 'c': 5}, {'a': 1, 'b': 4, 'c': 5}, {'a': 2, 'b': 3, 'c': 5}, {'a': 2, 'b': 4, 'c': 5}]

    References
    ----------
    https://stackoverflow.com/a/5228294/3010700
    """

    keys = kwargs.keys()
    for instance in itertools.product(*kwargs.values()):
        yield dict(zip(keys, instance, strict=True))


def update_dict_with_kwargs(dictionary, /, **kwargs):
    """update a dictionary with keyword arguments. The kwargs take precedence

    Parameters
    ----------
    dictionary : dict, optional
        The dictionary to update with the keyword arguments.
    **kwargs : keyword arguments
        The keyword arguments.

    Examples
    --------
    >>> update_dict_with_kwargs({"a": 1}, a=2)
    {'a': 2}
    >>> update_dict_with_kwargs({"a": 1, "b":2}, b=3, c=5)
    {'a': 1, 'b': 3, 'c': 5}
    """

    if not isinstance(dictionary, dict | type(None)):
        raise TypeError(
            f"First argument must be a dict or None, got '{type(dictionary).__name__}'"
        )

    return (dictionary or {}) | kwargs


def emit_user_level_warning(message, category=None) -> None:
    """emit a warning at the user level"""

    # skip_file_prefixes only defined in python 3.12
    if sys.version_info >= (3, 12):

        import filefisher

        pkg_dir = pathlib.Path(filefisher.__file__).parent.as_posix()

        warnings.warn(message, category=category, skip_file_prefixes=(pkg_dir,))

    else:

        warnings.warn(message, category=category, stacklevel=3)



import pandas as pd
import pytest

from filefisher import FileContainer
from filefisher.filters import priority_filter

from . import assert_no_warnings

# ({"model": ["a", "a", "b", "c", "c", "d"], "res": ["d", "h", "h", "h", "d", "z"]})


def test_priority_filter_errors():

    df = pd.DataFrame.from_records(
        [("a", "d")],
        columns=("model", "res"),
    )

    with pytest.raises(ValueError, match=r"column \('not_av'\) must be available"):
        priority_filter(df, "not_av", ["a"])

    with pytest.raises(ValueError, match=r"groupby` may not contain column \('res'\)"):
        priority_filter(df, "res", ["a"], groupby=["model", "res"])


def test_priority_filter_simple():

    df = pd.DataFrame.from_records(
        [("a", "d"), ("a", "h"), ("b", "h"), ("b", "d"), ("c", "d")],
        columns=("model", "res"),
    )

    expected = pd.DataFrame.from_records(
        [("a", "h"), ("b", "h"), ("c", "d")], columns=("model", "res"), index=(1, 2, 4)
    )

    result = priority_filter(df, "res", ["h", "d"])

    pd.testing.assert_frame_equal(result, expected)


def test_priority_filter_missing():

    df = pd.DataFrame.from_records(
        [("a", "d"), ("a", "h"), ("b", "d"), ("c", "z")], columns=("model", "res")
    )

    expected = pd.DataFrame.from_records(
        [("a", "h"), ("b", "d")], columns=("model", "res"), index=(1, 2)
    )

    with pytest.raises(
        ValueError, match="Did not find any element from the priority list"
    ):
        priority_filter(df, "res", ["h", "d"])

    with pytest.warns(
        UserWarning, match="Did not find any element from the priority list"
    ):
        result = priority_filter(df, "res", ["h", "d"], on_missing="warn")

    with pytest.raises(
        ValueError, match="Did not find any element from the priority list"
    ):
        with pytest.warns(
            FutureWarning, match="on_missing value 'error' has been renamed to 'raise'"
        ):
            result = priority_filter(df, "res", ["h", "d"], on_missing="error")

    with pytest.raises(
        ValueError,
        match="Unknown value for 'on_missing': 'foo'. Must be one of 'raise', 'warn' or 'ignore'.",
    ):
        result = priority_filter(df, "res", ["h", "d"], on_missing="foo")

    pd.testing.assert_frame_equal(result, expected)

    with assert_no_warnings():
        res = priority_filter(df, "res", ["h", "d"], on_missing="ignore")

    pd.testing.assert_frame_equal(res, expected)


def test_priority_filter_duplicates():

    df = pd.DataFrame.from_records([("a", "d"), ("a", "d")], columns=("model", "res"))

    with pytest.raises(ValueError, match="Found more than one"):
        priority_filter(df, "res", ["h", "d"])


def test_priority_filter_multi():

    df = pd.DataFrame.from_records(
        [("a", 1, "d"), ("a", 2, "d"), ("a", 1, "h"), ("b", 1, "h")],
        columns=("model", "number", "res"),
    )

    expected = df.iloc[[2, 1, 3]]

    result = priority_filter(df, "res", ["h", "d"])

    pd.testing.assert_frame_equal(result, expected)


def test_priority_filter_groupby():

    df = pd.DataFrame.from_records(
        [("a", 1, "d"), ("a", 2, "h"), ("b", 1, "h"), ("b", 2, "d")],
        columns=("model", "number", "res"),
    )

    result = priority_filter(df, "res", ["h", "d"])
    expected = df

    pd.testing.assert_frame_equal(result, expected)

    expected = pd.DataFrame.from_records(
        [("a", 2, "h"), ("b", 1, "h")], columns=("model", "number", "res")
    )
    result = priority_filter(df, "res", ["h", "d"], groupby=["model"])

    result = result.reset_index(drop=True)
    pd.testing.assert_frame_equal(result, expected)


def test_priority_filter_filename():
    """the filename is per default unique -> must be excluded from groupby"""
    df = pd.DataFrame.from_records(
        [
            ("file1", "a", "d"),
            ("file2", "a", "h"),
            ("file3", "b", "h"),
            ("file4", "b", "d"),
            ("file5", "c", "d"),
        ],
        columns=("path", "model", "res"),
    ).set_index("path")

    expected = pd.DataFrame.from_records(
        [("file2", "a", "h"), ("file3", "b", "h"), ("file5", "c", "d")],
        columns=("path", "model", "res"),
    ).set_index("path")

    result = priority_filter(df, "res", ["h", "d"])

    pd.testing.assert_frame_equal(result, expected)


def test_priority_filter_filecontainer_simple():
    """the filename is per default unique -> must be excluded from groupby"""
    df = pd.DataFrame.from_records(
        [
            ("file1", "a", "d"),
            ("file2", "a", "h"),
            ("file3", "b", "h"),
            ("file4", "b", "d"),
            ("file5", "c", "d"),
        ],
        columns=("path", "model", "res"),
    ).set_index("path")

    fc = FileContainer(df)

    expected = pd.DataFrame.from_records(
        [("file2", "a", "h"), ("file3", "b", "h"), ("file5", "c", "d")],
        columns=("path", "model", "res"),
    ).set_index("path")

    result = priority_filter(fc, "res", ["h", "d"])

    assert isinstance(result, FileContainer)

    pd.testing.assert_frame_equal(result.df, expected)



import pandas as pd

from filefisher import FileContainer
from filefisher.cmip import create_ensnumber, ensure_unique_grid, parse_ens


def test_parse_ens_cmip5():

    # pattern: "r{r}i{i}p{p}"

    df = pd.DataFrame.from_records(
        [
            ("file1", "r1i1p1"),
            ("file2", "r2i1p1"),
            ("file3", "r3i1p1"),
            ("file4", "r1i2p1"),
            ("file5", "r1i1p2"),
        ],
        columns=("path", "ens"),
    ).set_index("path")

    fc = FileContainer(df)

    expected_df = pd.DataFrame.from_records(
        [
            ("file1", "r1i1p1", "1", "1", "1"),
            ("file2", "r2i1p1", "2", "1", "1"),
            ("file3", "r3i1p1", "3", "1", "1"),
            ("file4", "r1i2p1", "1", "2", "1"),
            ("file5", "r1i1p2", "1", "1", "2"),
        ],
        columns=("path", "ens", "r", "i", "p"),
    ).set_index("path")

    result = parse_ens(fc)
    assert isinstance(result, FileContainer)

    pd.testing.assert_frame_equal(result.df, expected_df)


def test_parse_ens_cmip6():

    # pattern: "r{r}i{i}p{p}f{f}"

    df = pd.DataFrame.from_records(
        [
            ("file1", "r1i1p1f1"),
            ("file2", "r2i1p1f2"),
            ("file3", "r3i1p1f1"),
            ("file4", "r1i2p1f1"),
            ("file5", "r1i1p2f1"),
        ],
        columns=("path", "ens"),
    ).set_index("path")

    fc = FileContainer(df)

    expected_df = pd.DataFrame.from_records(
        [
            ("file1", "r1i1p1f1", "1", "1", "1", "1"),
            ("file2", "r2i1p1f2", "2", "1", "1", "2"),
            ("file3", "r3i1p1f1", "3", "1", "1", "1"),
            ("file4", "r1i2p1f1", "1", "2", "1", "1"),
            ("file5", "r1i1p2f1", "1", "1", "2", "1"),
        ],
        columns=("path", "ens", "r", "i", "p", "f"),
    ).set_index("path")

    result = parse_ens(fc)
    assert isinstance(result, FileContainer)

    pd.testing.assert_frame_equal(result.df, expected_df)


def test_create_ensnumber():
    columns = ("path", "model", "exp", "table", "varn", "ens")

    common = ("exp", "table", "varn")

    records = [
        ("file0", "CESM2", *common, "r1i1p1f1"),
        ("file1", "CESM2", *common, "r2i1p1f1"),
        ("file2", "CESM2", *common, "r3i1p1f1"),
        ("file3", "UKESM", *common, "r1i2p1f1"),
        ("file4", "UKESM", *common, "r1i3p1f1"),
    ]

    df = pd.DataFrame.from_records(
        records,
        columns=columns,
    ).set_index("path")

    expected_df = pd.DataFrame.from_records(
        records,
        columns=columns,
    ).set_index("path")
    expected_df["ensnumber"] = (0, 1, 2, 0, 1)

    fc = FileContainer(df)

    result = create_ensnumber(fc)
    assert isinstance(result, FileContainer)

    pd.testing.assert_frame_equal(result.df, expected_df)


def test_ensure_unique_grid():

    columns = ("path", "model", "exp", "table", "varn", "ens", "grid")

    # VALID_GRIDS = ("gn", "gr", "gr1", "gm")

    common = ("exp", "table", "varn", "ens")

    df = pd.DataFrame.from_records(
        [
            ("CESM2_gr", "CESM2", *common, "gr"),
            ("CESM2_gn", "CESM2", *common, "gn"),
            ("CESM2_gm", "CESM2", *common, "gm"),
            ("UKESM_gr", "UKESM", *common, "gr"),
            ("UKESM_gr1", "UKESM", *common, "gr1"),
        ],
        columns=columns,
    ).set_index("path")

    expected = pd.DataFrame.from_records(
        [
            ("CESM2_gn", "CESM2", *common, "gn"),
            ("UKESM_gr", "UKESM", *common, "gr"),
        ],
        columns=columns,
    ).set_index("path")

    result = ensure_unique_grid(df)

    pd.testing.assert_frame_equal(result, expected)



import sys

import pytest

from . import assert_no_warnings


def test_filefinder_deprecation():

    with pytest.warns(
        FutureWarning, match="`filefinder` has been renamed to `filefisher`"
    ):
        import filefinder  # noqa: F401

    # NOTE warning only issued once (even for from imports)
    with assert_no_warnings():
        from filefinder import FileContainer  # noqa: F401

    # remove imported module to re-trigger warning
    del sys.modules["filefinder"]

    with pytest.warns(
        FutureWarning, match="`filefinder` has been renamed to `filefisher`"
    ):
        from filefinder import FileFinder  # noqa: F401 # noqa: F401



import pandas as pd
import pytest

from filefisher import FileContainer

from . import assert_filecontainer_empty


@pytest.fixture
def example_df():

    df = pd.DataFrame.from_records(
        [
            ("file0", "a", "d", "r"),
            ("file1", "a", "h", "r"),
            ("file2", "b", "h", "r"),
            ("file3", "b", "d", "r"),
            ("file4", "c", "d", "r"),
        ],
        columns=("path", "model", "scen", "res"),
    ).set_index("path")

    return df


@pytest.fixture
def example_fc(example_df):

    return FileContainer(example_df)


def test_empty_filecontainer():

    df = pd.DataFrame([], columns=["cat"], index=pd.Index([], name="path"))

    fc = FileContainer(df)

    assert fc.df is df
    assert len(fc) == 0

    with pytest.raises(StopIteration):
        next(iter(fc.items()))


def test_filecontainer(example_df, example_fc):

    assert example_fc.df is example_df
    assert len(example_fc) == 5


def test_fc_iter(example_df, example_fc):

    # test one manually
    with pytest.warns(
        FutureWarning, match="iterating over a `FileContainer` is deprecated"
    ):
        path, meta = next(iter(example_fc))

    assert path == "file0"
    assert meta == {"model": "a", "scen": "d", "res": "r"}

    with pytest.warns(
        FutureWarning, match="iterating over a `FileContainer` is deprecated"
    ):
        result = list(example_fc)
    expected = list(
        zip(example_df.index.to_list(), example_df.to_dict("records"), strict=True)
    )
    assert result == expected


def test_fc_items(example_df, example_fc):

    # test one manually
    path, meta = next(iter(example_fc.items()))
    assert path == "file0"
    assert meta == {"model": "a", "scen": "d", "res": "r"}

    result = list(example_fc.items())
    expected = list(
        zip(example_df.index.to_list(), example_df.to_dict("records"), strict=True)
    )
    assert result == expected


def test_fc_getitem(example_df, example_fc):

    # indexing by scalar currently returns path, meta (analog to the loop)
    # not sure if this is the best way but keep for now
    path, meta = example_fc[0]
    assert path == "file0"
    assert meta == {"model": "a", "scen": "d", "res": "r"}

    path, meta = example_fc[-1]
    assert path == "file4"
    assert meta == {"model": "c", "scen": "d", "res": "r"}

    result = example_fc[0:1]
    assert isinstance(result, FileContainer)
    pd.testing.assert_frame_equal(result.df, example_df.iloc[0:1])

    result = example_fc[0:0]
    assert_filecontainer_empty(result, columns=["model", "scen", "res"])


def test_filecontainer_paths(example_fc):

    result = example_fc.paths
    expected = [
        "file0",
        "file1",
        "file2",
        "file3",
        "file4",
    ]

    assert result == expected


def test_filecontainer_meta(example_fc):

    result = example_fc.meta

    expected = [
        {"model": "a", "scen": "d", "res": "r"},
        {"model": "a", "scen": "h", "res": "r"},
        {"model": "b", "scen": "h", "res": "r"},
        {"model": "b", "scen": "d", "res": "r"},
        {"model": "c", "scen": "d", "res": "r"},
    ]

    assert result == expected


def test_filecontainer_search(example_df, example_fc):

    with pytest.raises(KeyError):
        example_fc.search(not_in_example_fc="a")

    assert_filecontainer_empty(example_fc.search())
    assert_filecontainer_empty(example_fc.search(model="d"))
    assert_filecontainer_empty(example_fc.search(model=0))

    result = example_fc.search(model="a")
    expected = example_df.iloc[[0, 1]]
    pd.testing.assert_frame_equal(result.df, expected)

    result = example_fc.search(model=["a", "c"])
    expected = example_df.iloc[[0, 1, 4]]
    pd.testing.assert_frame_equal(result.df, expected)

    result = example_fc.search(model="a", scen="h")
    expected = example_df.iloc[[1]]
    pd.testing.assert_frame_equal(result.df, expected)

    result = example_fc.search(model=["a", "b"], scen="d")
    expected = example_df.iloc[[0, 3]]
    pd.testing.assert_frame_equal(result.df, expected)

    # check if a tuple is accepted
    result = example_fc.search(model=("a", "b"), scen="d")
    expected = example_df.iloc[[0, 3]]
    pd.testing.assert_frame_equal(result.df, expected)


def test_filecontainer_search_none_key(example_df, example_fc):
    # https://github.com/mpytools/filefisher/issues/156

    result = example_fc.search(model=None)
    pd.testing.assert_frame_equal(result.df, example_df)

    result = example_fc.search(model="a", scen=None)
    expected = example_df.iloc[[0, 1]]
    pd.testing.assert_frame_equal(result.df, expected)


def test_filecontainer_search_single(example_df, example_fc):

    # empty search query results in an empty FileContainer
    with pytest.raises(ValueError, match="Found no paths"):
        example_fc.search_single()

    with pytest.raises(ValueError, match=r"Found more than one \(5\) paths"):
        example_fc.search_single(model=None)

    with pytest.raises(ValueError, match=r"Found more than one \(2\) paths"):
        example_fc.search_single(model="a")

    result = example_fc.search_single(model="a", scen="h")
    expected = example_df.iloc[[1]]
    pd.testing.assert_frame_equal(result.df, expected)


def test_filecontainer_concat(example_fc):

    with pytest.raises(ValueError, match="Can only concatenate two FileContainers."):
        example_fc.concat("not a FileContainer")

    example_fc_2 = FileContainer(example_fc.df)
    with pytest.raises(ValueError, match="FileContainers must have the same keys"):
        different_keys_fc = FileContainer(example_fc_2.df.loc[:, ["model", "scen"]])
        example_fc.concat(different_keys_fc)

    result = example_fc.concat(example_fc_2, drop_duplicates=False)
    expected = pd.concat([example_fc.df, example_fc_2.df])

    pd.testing.assert_frame_equal(result.df, expected)
    assert len(result) == 10

    result = example_fc.concat(example_fc_2, drop_duplicates=True)
    expected = example_fc

    pd.testing.assert_frame_equal(result.df, expected.df)
    assert len(result) == 5


def test_fc_combine_by_key_deprecated(example_fc):

    with pytest.warns(FutureWarning, match="`combine_by_key` has been deprecated"):
        example_fc[[0]].combine_by_key()


def test_fc_combine_by_keys(example_fc):

    result = example_fc[[0]]._combine_by_keys()

    # create one manually
    expected = pd.Series(["a.d.r"], index=pd.Index(["file0"], name="path"))
    pd.testing.assert_series_equal(result, expected)

    result = example_fc._combine_by_keys()
    expected = map(".".join, example_fc.df.values)
    expected = pd.Series(expected, index=example_fc.df.index)

    # different sep
    result = example_fc._combine_by_keys(sep="|")
    expected = map("|".join, example_fc.df.values)
    expected = pd.Series(expected, index=example_fc.df.index)

    # not all columns
    result = example_fc._combine_by_keys(keys=("model", "res"))
    expected = map(".".join, example_fc.df[["model", "res"]].values)
    expected = pd.Series(expected, index=example_fc.df.index)


def test_filefinder_repr(example_fc):

    # NOTE: does not test the pd.DataFrame part of the repr

    df = pd.DataFrame([], columns=["cat"], index=pd.Index([], name="path"))
    fc = FileContainer(df)
    assert fc.__repr__().startswith("<FileContainer: 0 paths>")

    assert example_fc.__repr__().startswith("<FileContainer: 5 paths>")



import pytest

from filefisher._utils import (
    _find_keys,
    atoi,
    natural_keys,
    product_dict,
    update_dict_with_kwargs,
)


@pytest.mark.parametrize(
    "string, expected",
    (
        ["/path/{var_name}/{year}", ("var_name", "year")],
        ["{b:d}{a:d}", ("b", "a")],
        ["{a:d}{b:d}", ("a", "b")],
        ["{a}{b:d}", ("a", "b")],
        ["{a:d}{b}", ("a", "b")],
        ["{a:d}{b:d}{c:d}", ("a", "b", "c")],
    ),
)
def test_find_keys(string, expected):

    result = _find_keys(string)

    assert result == expected


def test_atoi():

    assert atoi("10") == 10
    assert atoi("a10") == "a10"
    assert atoi("a") == "a"


def test_natural_keys_sort():

    lst = ["a10", "a1"]
    expected = ["a1", "a10"]
    assert not lst == expected

    lst.sort(key=natural_keys)
    assert lst == expected


def test_product_dict():

    result = list(product_dict(a=[1, 2], b=[3, 4], c=[5]))
    expected = [
        {"a": 1, "b": 3, "c": 5},
        {"a": 1, "b": 4, "c": 5},
        {"a": 2, "b": 3, "c": 5},
        {"a": 2, "b": 4, "c": 5},
    ]

    assert result == expected


def test_update_dict_with_kwargs():

    result = update_dict_with_kwargs({"a": 1}, a=2)
    expected = {"a": 2}
    assert result == expected

    result = update_dict_with_kwargs({"a": 1, "b": 2}, b=3, c=5)
    expected = {"a": 1, "b": 3, "c": 5}
    assert result == expected

    result = update_dict_with_kwargs({"a": 1, "b": 2})
    expected = {"a": 1, "b": 2}
    assert result == expected

    result = update_dict_with_kwargs(None, a=1, b=2)
    expected = {"a": 1, "b": 2}
    assert result == expected

    with pytest.raises(TypeError, match="First argument must be a dict"):
        update_dict_with_kwargs([])

    with pytest.raises(TypeError, match="got 'str'"):
        update_dict_with_kwargs("")

    with pytest.raises(TypeError, match="missing 1 required positional argument"):
        update_dict_with_kwargs(dictionary={})



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Import packages ---------------------------------------------------------

import datetime
import os
import subprocess
import sys
from importlib.metadata import version

import filefisher

# -- Display version info ----------------------------------------------------
# for debugging on RTD

on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if on_rtd:

    print("python exec:", sys.executable)
    print("sys.path:", sys.path)
    print("os.getcwd():", os.getcwd())

    if "CONDA_DEFAULT_ENV" in os.environ or "conda" in sys.executable:
        print("conda environment:")
        subprocess.run([os.environ.get("CONDA_EXE", "conda"), "list"])
    else:
        print("pip environment:")
        subprocess.run([sys.executable, "-m", "pip", "list"])

    print(f"filefisher: {filefisher.__version__=}, {filefisher.__file__=}")


# -- Project information -----------------------------------------------------

project = "filefisher"
copyright_year = datetime.date.today().year
copyright = f"(c) 2024-{copyright_year} Mathias Hauser"

authors = "Mathias Hauser"
author = authors

# The full version, including alpha/beta/rc tags
release = version("filefisher")
# The short X.Y version
version = ".".join(release.split(".")[:2])


# -- General configuration ---------------------------------------------------

# add sphinx extension modules
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

autosummary_generate = True

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_rtype = False

# napoleon_use_ivar = True
# napoleon_use_admonition_for_notes = True

numpydoc_class_members_toctree = True
numpydoc_show_class_members = False

autodoc_typehints = "none"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "sphinx_book_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = []

pygments_style = "sphinx"



import fnmatch
import glob
import logging
import os
import re
import warnings
from collections.abc import Generator
from typing import Any

import numpy as np
import pandas as pd
import parse

from filefisher._utils import (
    _find_keys,
    emit_user_level_warning,
    natural_keys,
    product_dict,
    update_dict_with_kwargs,
)

logger = logging.getLogger(__name__)

_FILE_FINDER_REPR = """<FileFinder{maybe_test_paths}>
path_pattern: '{path_pattern}'
file_pattern: '{file_pattern}'

keys: {repr_keys}
"""


def _deprecate_allow_empty(**kwargs):

    _allow_empty = kwargs.get("_allow_empty")

    if _allow_empty is not None:
        raise TypeError("`_allow_empty` has been deprecated in favour of `on_empty`")


_RESERVED_PLACEHOLDERS = {"keys", "on_parse_error", "on_empty", "_allow_empty"}


def _assert_valid_keys(keys) -> None:

    for key in _RESERVED_PLACEHOLDERS:
        if key in keys:
            raise ValueError(f"'{key}' is not a valid placeholder")


def _assert_unique(df) -> None:

    duplicates = df.duplicated()

    if duplicates.any():
        duplicated = df[duplicates].head()
        msg = f"Non-unique metadata detected.\nFirst five duplicates:\n{duplicated}"

        raise ValueError(msg)


class _FinderBase:
    def __init__(self, pattern, suffix=""):

        self.pattern = pattern
        self.keys = _find_keys(pattern)
        _assert_valid_keys(self.keys)
        self.parser = parse.compile(self.pattern)

        if self.parser.fixed_fields:
            msg = (
                "Only named fields are currently allowed: avoid empty braces and"
                " leading underscores."
            )
            raise ValueError(msg)

        self._suffix = suffix

        # replace the fmt spec - add the capture group again
        self._pattern_no_fmt_spec = re.sub(
            r"\{([A-Za-z0-9_]+)(:.*?)\}", r"{\1}", pattern
        )

    def create_name(self, keys=None, **keys_kwargs) -> str:
        """build name from keys

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the name.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.
        """

        keys = update_dict_with_kwargs(keys, **keys_kwargs)

        return self._pattern_no_fmt_spec.format(**keys)


class _Finder(_FinderBase):
    def _create_condition_dict(self, **kwargs):

        # add wildcard for all undefined keys
        cond_dict = dict.fromkeys(self.keys, "*")
        cond_dict.update(**kwargs)

        return cond_dict

    def find(
        self, keys=None, *, on_parse_error="raise", on_empty="raise", **keys_kwargs
    ) -> "FileContainer":
        """find files in the file system using the file and path (folder) pattern

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the search pattern. Several names can
            be passed for each key as list (interpreted as or). None is ignored.
        on_parse_error : "raise" | "warn" | "ignore", default: "raise"
            What to do if a path/file name cannot be parsed. If "raise" raises a ValueError,
            if "warn" raises a warning and if "ignore" ignores the file.
        on_empty : "raise" | "warn" | "allow", default: "raise"
            Behaviour when no files are found: "raise" (default) raises a ValueError,
            "warn" raises a warning. For "warn" and "allow" an empty FileContainer is returned.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Notes
        -----
        Missing ``keys`` are replaced with ``"*"``.

        """

        _deprecate_allow_empty(**keys_kwargs)

        keys = update_dict_with_kwargs(keys, **keys_kwargs)

        # ignore None values
        keys = {key: value for key, value in keys.items() if value is not None}

        if on_parse_error not in ("raise", "warn", "ignore"):
            raise ValueError(
                f"Unknown value for 'on_parse_error': '{on_parse_error}'. Must be one of 'raise', 'warn' or 'ignore'."
            )

        if on_empty not in ("raise", "warn", "allow"):
            raise ValueError(
                f"Unknown value for 'on_empty': '{on_empty}'. Must be one of 'raise', 'warn' or 'allow'."
            )

        # wrap strings and scalars in list
        for key, value in keys.items():
            if isinstance(value, str) or np.ndim(value) == 0:
                keys[key] = [value]

        all_paths = list()
        all_patterns = list()
        for one_search_dict in product_dict(**keys):

            cond_dict = self._create_condition_dict(**one_search_dict)
            full_pattern = self.create_name(**cond_dict)

            paths = sorted(self._glob(full_pattern), key=natural_keys)

            all_paths += paths

            all_patterns.append(full_pattern)

        if len(all_paths) == 0:
            msg = "Found no files matching criteria. Tried the following pattern(s):"
            msg += "".join(f"\n- '{pattern}'" for pattern in all_patterns)

            if on_empty == "raise":
                raise ValueError(msg)
            elif on_empty == "warn":
                emit_user_level_warning(msg)

        # NOTE: also creates the correct (empty) df if no paths are found
        df = self._parse_paths(all_paths, on_parse_error=on_parse_error)
        _assert_unique(df)

        return FileContainer(df)

    def find_single(self, keys=None, **keys_kwargs) -> "FileContainer":
        """
        find exactly one file/ path in the file system using the file and path pattern

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the search pattern.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Notes
        -----
        Missing ``keys`` are replaced with ``"*"``.

        Raises
        ------
        ValueError : if more or less than one file/ path is found
        """

        fc = self.find(keys, on_parse_error="raise", on_empty="raise", **keys_kwargs)

        if len(fc) > 1:
            n_found = len(fc)
            msg = (
                f"Found more than one ({n_found}) files/ paths. Please adjust your"
                f" query.\nFirst five files/ paths:\n{fc.df.head()}"
            )
            raise ValueError(msg)

        return fc

    @staticmethod
    def _glob(pattern) -> list[str]:
        """Return a list of paths matching a pathname pattern

        Notes
        -----
        glob has it's own method so it can be mocked by the tests

        """

        return glob.glob(pattern)

    def _parse_paths(self, paths, on_parse_error) -> pd.DataFrame:

        valid_paths, out = list(), list()
        for path in paths:
            parsed = self.parser.parse(path)

            if parsed is None:
                if on_parse_error == "raise":
                    raise ValueError(
                        f"Could not parse '{path}' with the pattern '{self.pattern}' - are"
                        " there contradictory values?"
                    )
                elif on_parse_error == "warn":
                    emit_user_level_warning(
                        f"Could not parse '{path}' with the pattern '{self.pattern}' - are"
                        " there contradictory values?"
                    )
                elif on_parse_error == "ignore":
                    pass
            else:
                valid_paths.append(path)
                out.append(list(parsed.named.values()))

        index = pd.Index(valid_paths, name="path") + self._suffix
        df = pd.DataFrame(out, columns=self.keys, index=index)
        return df


class FileFinder:

    def __init__(
        self, path_pattern: str, file_pattern: str, *, test_paths=None
    ) -> None:
        """find and create file names based on python format syntax

        Parameters
        ----------
        path_pattern : str
            String denoting the path (folder) pattern where everything variable is
            enclosed in curly braces.
        file_pattern : str
            String denoting the file pattern where everything variable is enclosed in
            curly braces.
        test_paths : list of str, default None
            A list of paths to use instead of querying the file system. To be used for
            testing and demonstration.

        Examples
        --------
        >>> path_pattern = "/root/{category}"
        >>> file_pattern = "{category}_file_{number}"

        >>> ff = FileFinder(path_pattern, file_pattern)
        """

        if os.path.sep in file_pattern:
            raise ValueError(
                f"`file_pattern` cannot contain path separator ('{os.path.sep}')"
            )

        # cannot search for files (only paths and full)
        self.file = _FinderBase(file_pattern)
        # ensure path_pattern ends with a /
        self.path = _Finder(os.path.join(path_pattern, ""), suffix="*")
        self.full = _Finder(os.path.join(*filter(None, (path_pattern, file_pattern))))

        self.keys_path = self.path.keys
        self.keys_file = self.file.keys
        self.keys = self.full.keys

        self.file_pattern = self.file.pattern
        self.path_pattern = self.path.pattern
        self._full_pattern = self.full.pattern

        if test_paths is not None:

            if isinstance(test_paths, str):
                test_paths = [test_paths]

            if len(test_paths) != len(set(test_paths)):
                raise ValueError("`test_paths` are not unique")

            self._set_test_paths(test_paths)
        else:
            self._test_paths = None

    def _set_test_paths(self, test_paths):

        self._test_paths = test_paths

        # use fnmatch.filter to 'glob' pseudo-filenames
        def finder(pat):

            # make fnmatch work (almost) the same as glob
            if pat.endswith(os.path.sep):
                # remove duplicate paths (i.e. if the filename made it non-unique)
                paths_ = sorted(
                    set(os.path.dirname(s) + os.path.sep for s in test_paths)
                )
            else:
                paths_ = test_paths

            return fnmatch.filter(paths_, pat)

        # overwrite the glob implementation
        self.path._glob = finder
        self.full._glob = finder

    def create_path_name(self, keys=None, **keys_kwargs) -> str:
        """build path (folder) name from keys

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the path (folder) name.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Examples
        --------
        >>> path_pattern = "/root/{category}"
        >>> file_pattern = "{category}_file_{number}"

        >>> ff = FileFinder(path_pattern, file_pattern)
        >>> ff.create_path_name(category="foo")
        '/root/foo/'

        >>> ff.create_path_name(dict(category="foo"))
        '/root/foo/'

        >>> ff.create_path_name(dict(category="foo"), category="bar")
        '/root/bar/'
        """

        # warnings.warn("'create_path_name' is deprecated, use 'path.name' instead")
        return self.path.create_name(keys, **keys_kwargs)

    def create_file_name(self, keys=None, **keys_kwargs) -> str:
        """build file name from keys

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the file name.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Examples
        --------
        >>> path_pattern = "/root/{category}"
        >>> file_pattern = "{category}_file_{number}"

        >>> ff = FileFinder(path_pattern, file_pattern)
        >>> ff.create_file_name(category="foo", number=1)
        'foo_file_1'

        >>> ff.create_file_name(dict(category="foo", number=1))
        'foo_file_1'

        >>> ff.create_file_name(dict(category="foo", number=1), category="bar")
        'bar_file_1'
        """

        # warnings.warn("'create_file_name' is deprecated, use 'file.name' instead")
        return self.file.create_name(keys, **keys_kwargs)

    def create_full_name(self, keys=None, **keys_kwargs) -> str:
        """build full (folder + file) name from keys

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the full (folder + file) name.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Examples
        --------
        >>> path_pattern = "/root/{category}"
        >>> file_pattern = "{category}_file_{number}"

        >>> ff = FileFinder(path_pattern, file_pattern)
        >>> ff.create_full_name(category="foo", number=1)
        '/root/foo/foo_file_1'

        >>> ff.create_full_name(dict(category="foo", number=1))
        '/root/foo/foo_file_1'

        >>> ff.create_full_name(dict(category="foo", number=1), category="bar")
        '/root/bar/bar_file_1'
        """

        # warnings.warn("'create_full_name' is deprecated, use 'full.name' instead")
        return self.full.create_name(keys, **keys_kwargs)

    def find_paths(
        self, keys=None, *, on_parse_error="raise", on_empty="raise", **keys_kwargs
    ) -> "FileContainer":
        """find files in the file system using the file and path (folder) pattern

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the search pattern. Several names can
            be passed for each key as list (interpreted as or). None is ignored.
        on_parse_error : "raise" | "warn" | "skip", default: "raise"
            What to do if a path/file name cannot be parsed. If "raise" raises a ValueError,
            if "warn" raises a warning and if "skip" ignores the file.
        on_empty : "raise" | "warn" | "allow", default: "raise"
            Behaviour when no files are found: "raise" (default) raises a ValueError,
            "warn" raises a warning. For "warn" and "allow" an empty FileContainer is returned.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Notes
        -----
        Missing ``keys`` are replaced with ``"*"``.

        Examples
        --------
        >>> path_pattern = "/root/{category}"
        >>> file_pattern = "{category}_file_{number}"
        >>> ff = FileFinder(path_pattern, file_pattern)

        >>> ff.find_paths()   # doctest: +SKIP
        Looks for
        - "/root/*/"

        >>> ff.find_paths(category="foo")   # doctest: +SKIP
        Looks for
        - "/root/foo/"

        >>> ff.find_paths(dict(category=["foo", "bar"]))   # doctest: +SKIP
        Looks for
        - "/root/foo/"
        - "/root/bar/"
        """
        return self.path.find(
            keys,
            on_parse_error=on_parse_error,
            on_empty=on_empty,
            **keys_kwargs,
        )

    def find_files(
        self, keys=None, *, on_parse_error="raise", on_empty="raise", **keys_kwargs
    ) -> "FileContainer":
        """find files in the file system using the file pattern

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the search pattern. Several names can
            be passed for each key as list (interpreted as or). None is ignored.
        on_parse_error : "raise" | "warn" | "skip", default: "raise"
            What to do if a path/file name cannot be parsed. If "raise" raises a ValueError,
            if "warn" raises a warning and if "skip" ignores the file.
        on_empty : "raise" | "warn" | "allow", default: "raise"
            Behaviour when no files are found: "raise" (default) raises a ValueError,
            "warn" raises a warning. For "warn" and "allow" an empty FileContainer is returned.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Notes
        -----
        Missing ``keys`` are replaced with ``"*"``.

        Examples
        --------
        >>> path_pattern = "/root/{category}"
        >>> file_pattern = "{category}_file_{number}"
        >>> ff = FileFinder(path_pattern, file_pattern)

        >>> ff.find_files()   # doctest: +SKIP
        Looks for
        - "/root/*/*_file_*"

        >>> ff.find_files(number=[1, 2])   # doctest: +SKIP
        Looks for
        - "/root/*/*_file_1"
        - "/root/*/*_file_2"

        >>> meta = {"category": "foo", "number": [1, 2]}
        >>> ff.find_files(meta, category="bar")   # doctest: +SKIP
        Looks for
        - "/root/bar/bar_file_1"
        - "/root/bar/bar_file_2"

        """
        return self.full.find(
            keys,
            on_parse_error=on_parse_error,
            on_empty=on_empty,
            **keys_kwargs,
        )

    def find_single_path(self, keys=None, **keys_kwargs) -> "FileContainer":
        """
        find exactly one path in the file system using the path pattern

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the search pattern.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Notes
        -----
        Missing ``keys`` are replaced with ``"*"``.

        Raises
        ------
        ValueError
            if more or less than one path is found.
        """

        return self.path.find_single(keys, **keys_kwargs)

    def find_single_file(self, keys=None, **keys_kwargs) -> "FileContainer":
        """
        find exactly one file in the file system using the file and path pattern

        Parameters
        ----------
        keys : dict
            Dictionary containing keys to create the search pattern.
        **keys_kwargs : {key: indexer, ...}, optional
            The keyword arguments form of ``keys``. When the same key is passed in
            ``keys`` and ``keys_kwargs`` the latter takes priority.

        Notes
        -----
        Missing ``keys`` are replaced with ``"*"``.

        Raises
        ------
        ValueError
            if more or less than one file is found.
        """

        return self.full.find_single(keys, **keys_kwargs)

    def __repr__(self) -> str:

        repr_keys = "', '".join(sorted(self.full.keys))
        repr_keys = f"'{repr_keys}'"

        tp = self._test_paths
        maybe_test_paths = "" if tp is None else f" ({len(tp)} test_paths)"

        msg = _FILE_FINDER_REPR.format(
            maybe_test_paths=maybe_test_paths,
            path_pattern=self.path.pattern,
            file_pattern=self.file.pattern,
            repr_keys=repr_keys,
        )

        return msg


class FileContainer:

    def __init__(self, df: pd.DataFrame):
        """FileContainer gathers paths and their metadata

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame with info about found paths from FileFinder.
        """

        self.df = df

    def __iter__(self):

        warnings.warn(
            "iterating over a `FileContainer` is deprecated. Use ``.paths``, ``.meta``"
            " or ``.items() instead.",
            FutureWarning,
            stacklevel=2,
        )

        for index, element in self.df.iterrows():
            yield index, element.to_dict()

    def __getitem__(self, key):

        if isinstance(key, int | np.integer):
            # use iloc -> there can be more than one element with index 0
            element = self.df.iloc[key]

            return element.name, element.to_dict()
        # assume slice or [1]
        else:
            df = self.df.iloc[key]
            return type(self)(df)

    @property
    def meta(self) -> list[dict[str, Any]]:
        """Return metadata as list of dictionaries"""
        return self.df.to_dict("records")

    @property
    def paths(self) -> list[str]:
        """Return paths as list"""
        return self.df.index.to_list()

    def items(self) -> Generator[tuple[str, dict[str, Any]], None, None]:
        """Return a generator of (path, metadata) tuples"""
        for index, element in self.df.iterrows():
            yield index, element.to_dict()

    def combine_by_key(self, keys=None, sep="."):
        """combine columns

        Parameters
        ----------
        keys : list[str], optional
            List of keys to combine. If None, all keys are combined.
        sep : str, default "."
            Separator between the keys.

        Returns
        -------
        pd.Series
            pd.Series with combined columns where the keys are separated by `sep`.

        """
        emit_user_level_warning(
            "`combine_by_key` has been deprecated and will be removed in a future version",
            FutureWarning,
        )

        return self._combine_by_keys(keys=keys, sep=sep)

    def _combine_by_keys(self, keys=None, sep="."):
        """combine columns

        Parameters
        ----------
        keys : list[str], optional
            List of keys to combine. If None, all keys are combined.
        sep : str, default "."
            Separator between the keys.

        Returns
        -------
        pd.Series
            pd.Series with combined columns where the keys are separated by `sep`.

        """

        if keys is None:
            keys = list(self.df.columns)

        return self.df[list(keys)].apply(lambda x: sep.join(x.map(str)), axis=1)

    def search(self, **query) -> "FileContainer":
        """subset paths given a search query

        Parameters
        ----------
        **query : Mapping[str, str | int | list[str | int]]
            Search query.

        Returns
        -------
        search_result : FileContainer

        Notes
        -----
        - individual conditions are combined with "and", e.g., ``model="a", exp="b"``
          requires the model to be "a" and the experiment to be "b".
        - conditions for a key are combined with "or", e.g., ``model=["a", "b"]``
          matches for both.
        - passing ``None`` as condition ignores it, e.g., ``model=None``.
        """

        df = self._get_subset(**query)
        return type(self)(df)

    def search_single(self, **query) -> "FileContainer":
        """find exactly one path given a search query - raises an error otherwise

        Parameters
        ----------
        **query : Mapping[str, str | int | list[str | int]]
            Search query.

        Returns
        -------
        search_result : FileContainer

        See Also
        --------
        FileContainer.search
        """

        fc = self.search(**query)

        if len(fc) == 0:
            msg = "Found no paths"
            raise ValueError(msg)

        if len(fc) > 1:
            n_found = len(fc)
            msg = (
                f"Found more than one ({n_found}) paths. Please adjust your"
                f" query.\nFirst five paths:\n{fc.df.head()}"
            )
            raise ValueError(msg)

        return fc

    def concat(self, other, drop_duplicates=True):
        """concatenate two FileContainers

        Parameters
        ----------
        other : FileContainer
            The other FileContainer to concatenate.
        drop_duplicates : bool, default True
            Drop duplicates after concatenating.

        Returns
        -------
        FileContainer
            The concatenated FileContainer.

        Raises
        ------
        ValueError
            If the other object is not a FileContainer.
        ValueError
            If the two FileContainers do not have the same keys (in the same order).
        """

        if not isinstance(other, FileContainer):
            raise ValueError("Can only concatenate two FileContainers.")

        if not self.df.columns.equals(other.df.columns):
            raise ValueError("FileContainers must have the same keys.")

        df = pd.concat([self.df, other.df])

        if drop_duplicates:
            df = df.drop_duplicates()

        return type(self)(df)

    def _get_subset(self, **query):
        if not query:
            return pd.DataFrame(
                [], columns=self.df.columns, index=pd.Index([], name="path")
            )

        # ignore None values
        query = {key: value for key, value in query.items() if value is not None}

        sel = np.ones_like(self.df.index, dtype=bool)
        for key, value in query.items():
            # isin does not handle scalars
            value = [value] if np.ndim(value) == 0 else value

            sel &= self.df[key].isin(value)

        return self.df[sel]

    def __len__(self):
        return self.df.__len__()

    def __repr__(self):

        n_paths = len(self)

        msg = f"<FileContainer: {n_paths} paths>\n"
        return msg + self.df.__repr__()



import warnings
from collections.abc import Iterable
from contextlib import contextmanager

import pandas as pd

from filefisher import FileContainer


@contextmanager
def assert_no_warnings():

    with warnings.catch_warnings(record=True) as record:
        yield record
        assert len(record) == 0, "got unexpected warning(s)"


def assert_filecontainer_empty(
    fc: FileContainer, columns: str | Iterable[str] | None = None
):
    """checks passed object is an empty `FileContainer`

    Parameters
    ----------
    fc : FileContainer
        Filecontainer to ensure is empty.
    columns : str, iterable of str | None, optional
        Keys/ columns the empty FileContainer should have.
    """

    assert isinstance(fc, FileContainer)

    assert isinstance(fc.df, pd.DataFrame)
    assert fc.df.index.name == "path"
    assert len(fc) == 0, f"FileContainer not empty ({len(fc)=})"

    if isinstance(columns, str):
        columns = {columns}

    # DataFrame should contain the keys as empty columns
    if columns:
        assert set(fc.df.columns) == set(columns)



import os
import textwrap

import pandas as pd
import pytest

from filefisher import FileFinder
from filefisher._filefinder import _assert_unique

from . import assert_filecontainer_empty


@pytest.fixture(scope="module")
def tmp_path(tmp_path_factory):
    return tmp_path_factory.mktemp("filefisher")


@pytest.fixture(scope="module", autouse=True)
def path(tmp_path):
    """
    Creates the following temporary structure:
    - /tmp/filefisher/a1/foo/file
    - /tmp/filefisher/a2/foo/file
    """

    d = tmp_path / "a1" / "foo"
    d.mkdir(parents=True)
    f = d / "file"
    f.write_text("")

    d = tmp_path / "a2" / "foo"
    d.mkdir(parents=True)
    f = d / "file"
    f.write_text("")

    return tmp_path


@pytest.fixture(scope="module", params=["from_filesystem", "from_string"])
def test_paths(request, tmp_path):

    if request.param == "from_filesystem":
        return None

    paths = ["a1/foo/file", "a2/foo/file"]
    paths = [str(tmp_path / path) for path in paths]

    return paths


@pytest.mark.parametrize("placeholder", ("keys", "on_parse_error", "_allow_empty"))
def test_pattern_invalid_placeholder(placeholder):

    with pytest.raises(ValueError, match=f"'{placeholder}' is not a valid placeholder"):
        FileFinder("", f"{{{placeholder}}}")

    with pytest.raises(ValueError, match=f"'{placeholder}' is not a valid placeholder"):
        FileFinder(f"{{{placeholder}}}", "")


@pytest.mark.parametrize("pattern", ("{}", "{_fixed}"))
def test_only_named_fields(pattern):

    with pytest.raises(ValueError, match="Only named fields are currently allowed"):
        FileFinder("", pattern)

    with pytest.raises(ValueError, match="Only named fields are currently allowed"):
        FileFinder(pattern, "")


def test_assert_unique():

    # no error raised
    df = pd.DataFrame.from_records([("a", "d"), ("a", "h")], columns=("model", "res"))
    _assert_unique(df)

    df = pd.DataFrame.from_records([("a", "d"), ("a", "d")], columns=("model", "res"))
    with pytest.raises(ValueError, match="Non-unique metadata detected"):
        _assert_unique(df)


@pytest.mark.parametrize("_allow_empty", (False, True))
def test_deprecate_allow_empty(_allow_empty):

    ff = FileFinder("", "a")
    msg = "`_allow_empty` has been deprecated in favour of `on_empty`"
    with pytest.raises(TypeError, match=msg):
        ff.find_files(_allow_empty=_allow_empty)

    with pytest.raises(TypeError, match=msg):
        ff.find_paths(_allow_empty=_allow_empty)


def test_wrong_on_empty():

    ff = FileFinder("", "a")
    msg = "Unknown value for 'on_empty': 'null'. Must be one of 'raise', 'warn' or 'allow'."
    with pytest.raises(ValueError, match=msg):
        ff.find_paths(on_empty="null")

    with pytest.raises(ValueError, match=msg):
        ff.find_files(on_empty="null")


def test_pattern_property():

    path_pattern = "path_pattern/"
    file_pattern = "file_pattern"

    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    assert ff.path_pattern == path_pattern
    assert ff.file_pattern == file_pattern

    assert ff._full_pattern == path_pattern + file_pattern

    assert ff.path.pattern == path_pattern
    assert ff.file.pattern == file_pattern

    assert ff.full.pattern == path_pattern + file_pattern


def test_test_path_property():

    ff = FileFinder("a", "b")

    assert ff._test_paths is None

    ff = FileFinder("a", "b", test_paths="path")
    assert ff._test_paths == ["path"]

    ff = FileFinder("a", "b", test_paths=["a", "b"])
    assert ff._test_paths == ["a", "b"]


def test_test_path_assert_nonunique():

    with pytest.raises(ValueError, match="`test_paths` are not unique"):
        FileFinder("a", "", test_paths=["a", "a"])

    with pytest.raises(ValueError, match="`test_paths` are not unique"):
        FileFinder("a", "", test_paths=["a/", "a/"])

    with pytest.raises(ValueError, match="`test_paths` are not unique"):
        FileFinder("a", "", test_paths=["a/b", "a/b"])


def test_file_pattern_no_sep():

    path_pattern = "path_pattern"
    file_pattern = "file" + os.path.sep + "pattern"

    with pytest.raises(ValueError, match="cannot contain path separator"):
        FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)


def test_pattern_sep_added():

    path_pattern = "path_pattern"
    file_pattern = "file_pattern"

    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)
    assert ff.path_pattern == path_pattern + os.path.sep


def test_keys():

    file_pattern = "{a}_{b}_{c}"
    path_pattern = "{ab}_{c}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    expected = ("ab", "c", "a", "b")
    assert ff.keys == expected

    expected = ("a", "b", "c")
    assert ff.keys_file == expected

    expected = ("ab", "c")
    assert ff.keys_path == expected


def test_repr():

    path_pattern = "/{a}/{b}"
    file_pattern = "{b}_{c}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    expected = """\
    <FileFinder>
    path_pattern: '/{a}/{b}/'
    file_pattern: '{b}_{c}'

    keys: 'a', 'b', 'c'
    """
    expected = textwrap.dedent(expected)
    assert expected == ff.__repr__()

    path_pattern = "{a}"
    file_pattern = "file_pattern"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    expected = """\
    <FileFinder>
    path_pattern: '{a}/'
    file_pattern: 'file_pattern'

    keys: 'a'
    """
    expected = textwrap.dedent(expected)
    assert expected == ff.__repr__()


def test_repr_test_paths():

    path_pattern = "/{a}/{b}"
    file_pattern = "{b}_{c}"
    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=["a"]
    )

    expected = """\
    <FileFinder (1 test_paths)>
    path_pattern: '/{a}/{b}/'
    file_pattern: '{b}_{c}'

    keys: 'a', 'b', 'c'
    """
    expected = textwrap.dedent(expected)

    print(expected)

    print(ff.__repr__())

    assert expected == ff.__repr__()

    path_pattern = "{a}"
    file_pattern = "file_pattern"
    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=["a", "b"]
    )

    expected = """\
    <FileFinder (2 test_paths)>
    path_pattern: '{a}/'
    file_pattern: 'file_pattern'

    keys: 'a'
    """
    expected = textwrap.dedent(expected)
    assert expected == ff.__repr__()


def test_create_name():

    path_pattern = "{a}/{b}"
    file_pattern = "{b}_{c}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    result = ff.create_path_name(a="a", b="b")
    assert result == "a/b/"

    result = ff.create_file_name(b="b", c="c")
    assert result == "b_c"

    result = ff.create_full_name(a="a", b="b", c="c")
    assert result == "a/b/b_c"


def test_create_name_dict():

    path_pattern = "{a}/{b}"
    file_pattern = "{b}_{c}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    result = ff.create_path_name(dict(a="a", b="b"))
    assert result == "a/b/"

    result = ff.create_file_name(dict(b="b", c="c"))
    assert result == "b_c"

    result = ff.create_full_name(dict(a="a", b="b", c="c"))
    assert result == "a/b/b_c"


def test_create_name_kwargs_priority():

    path_pattern = "{a}/{b}"
    file_pattern = "{b}_{c}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    result = ff.create_path_name(dict(a="XXX", b="b"), a="a")
    assert result == "a/b/"

    result = ff.create_file_name(dict(b="XXX", c="c"), b="b")
    assert result == "b_c"

    result = ff.create_full_name(dict(a="XXX", b="b"), a="a", c="c")
    assert result == "a/b/b_c"


def test_find_path_none_found(tmp_path, test_paths):

    path_pattern = tmp_path / "{a}/foo/"
    file_pattern = "file_pattern"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    with pytest.raises(ValueError, match="Found no files matching criteria"):
        ff.find_paths(a="foo")

    with pytest.raises(ValueError, match="Found no files matching criteria"):
        ff.find_paths({"a": "foo"})

    with pytest.warns(match="Found no files matching criteria"):
        result = ff.find_paths(a="foo", on_empty="warn")
    assert_filecontainer_empty(result, columns="a")

    with pytest.warns(match="Found no files matching criteria"):
        result = ff.find_paths({"a": "foo"}, on_empty="warn")
    assert_filecontainer_empty(result, columns="a")

    result = ff.find_paths(a="foo", on_empty="allow")
    assert_filecontainer_empty(result, columns="a")

    result = ff.find_paths({"a": "foo"}, on_empty="allow")
    assert_filecontainer_empty(result, columns="a")


def test_find_paths_non_unique():

    # ensure find_paths works for duplicated folder names (made unique by the file name)

    ff = FileFinder("{cat}", "", test_paths=["a/a", "a/b"])

    result = ff.find_paths()

    expected = {"path": {0: "a/*"}, "cat": {0: "a"}}
    expected = pd.DataFrame.from_dict(expected).set_index("path")
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_paths_simple(tmp_path, test_paths):

    path_pattern = tmp_path / "a1/{a}/"
    file_pattern = "file_pattern"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {"path": {0: str(tmp_path / "a1/foo/*")}, "a": {0: "foo"}}
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_paths(a="foo")
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_paths(dict(a="foo"))
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_paths(dict(a="XXX"), a="foo")
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_paths_none_key_ignored():

    ff = FileFinder("{folder}", "", test_paths=["a/"])

    expected = {"path": {0: "a/*"}, "folder": {0: "a"}}
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_paths(path=None)
    pd.testing.assert_frame_equal(result.df, expected)


@pytest.mark.parametrize("find_kwargs", [{"b": "foo"}, {"a": "*", "b": "foo"}])
def test_find_paths_wildcard(tmp_path, test_paths, find_kwargs):

    path_pattern = tmp_path / "{a}/{b}"
    file_pattern = "file_pattern"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {
        "path": {0: str(tmp_path / "a1/foo/*"), 1: str(tmp_path / "a2/foo/*")},
        "a": {0: "a1", 1: "a2"},
        "b": {0: "foo", 1: "foo"},
    }
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_paths(**find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_paths(find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_paths({"b": "XXX"}, **find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)


@pytest.mark.parametrize(
    "find_kwargs",
    [{"a": ["a1", "a2"], "b": "foo"}, {"a": ["a1", "a2"], "b": ["foo", "bar"]}],
)
def test_find_paths_several(tmp_path, test_paths, find_kwargs):

    path_pattern = tmp_path / "{a}/{b}"
    file_pattern = "file_pattern"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {
        "path": {0: str(tmp_path / "a1/foo/*"), 1: str(tmp_path / "a2/foo/*")},
        "a": {0: "a1", 1: "a2"},
        "b": {0: "foo", 1: "foo"},
    }
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_paths(**find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_paths(find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_paths({"a": "XXX", "b": "XXX"}, **find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)


@pytest.mark.parametrize(
    "find_kwargs",
    [{"a": "a1"}, {"a": "a1", "b": "foo"}, {"a": "a1", "b": ["foo", "bar"]}],
)
def test_find_paths_one_of_several(tmp_path, test_paths, find_kwargs):

    path_pattern = tmp_path / "{a}/{b}"
    file_pattern = "file_pattern"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {
        "path": {0: str(tmp_path / "a1/foo/*")},
        "a": {0: "a1"},
        "b": {0: "foo"},
    }
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_paths(**find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_paths(find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_paths({"a": "XXX"}, **find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_single_path(tmp_path, test_paths):

    path_pattern = tmp_path / "{a}/foo"
    file_pattern = "file_pattern"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    # error if more than one is found
    with pytest.raises(ValueError, match=r"Found more than one \(2\) files/ paths"):
        ff.find_single_path()

    # error if more than one is found
    with pytest.raises(ValueError, match="Found no files matching criteria"):
        ff.find_single_path(a="a3")

    expected = {"path": {0: str(tmp_path / "a1/foo/*")}, "a": {0: "a1"}}
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_single_path(a="a1")
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_single_path({"a": "a1"})
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_file_none_found(tmp_path, test_paths):

    path_pattern = tmp_path / "{a}/foo/"
    file_pattern = "{file_pattern}"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    with pytest.raises(ValueError, match="Found no files matching criteria"):
        ff.find_files(a="XXX")

    with pytest.raises(ValueError, match="Found no files matching criteria"):
        ff.find_files({"a": "XXX"})

    with pytest.warns(match="Found no files matching criteria"):
        result = ff.find_files(a="XXX", on_empty="warn")
    assert_filecontainer_empty(result, columns=("a", "file_pattern"))

    with pytest.warns(match="Found no files matching criteria"):
        result = ff.find_files({"a": "XXX"}, on_empty="warn")
    assert_filecontainer_empty(result, columns=("a", "file_pattern"))

    result = ff.find_files(a="XXX", on_empty="allow")
    assert_filecontainer_empty(result, columns=("a", "file_pattern"))

    result = ff.find_files({"a": "XXX"}, on_empty="allow")
    assert_filecontainer_empty(result, columns=("a", "file_pattern"))

    result = ff.find_files({"a": "XXX"}, on_empty="allow", a="XXX")
    assert_filecontainer_empty(result, columns=("a", "file_pattern"))


def test_find_files_non_unique():

    # test raises error for non-unique metadata - AFAIK not possible for real paths

    ff = FileFinder("", "{cat}")
    # need to set via `_set_test_paths` to avoid duplicate check
    ff._set_test_paths(test_paths=["/a", "/a"])

    with pytest.raises(ValueError, match="Non-unique metadata detected"):
        ff.find_files()


def test_find_file_simple(tmp_path, test_paths):

    path_pattern = tmp_path / "a1/{a}/"
    file_pattern = "file"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {"path": {0: str(tmp_path / "a1/foo/file")}, "a": {0: "foo"}}
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_files(a="foo")
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files({"a": "foo"})
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files({"a": "XXX"}, a="foo")
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_file_none_key_ignored():

    ff = FileFinder("", "{file}", test_paths=["a"])

    expected = {"path": {0: "a"}, "file": {0: "a"}}
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_files(file=None)

    pd.testing.assert_frame_equal(result.df, expected)


@pytest.mark.parametrize("find_kwargs", [{"b": "file"}, {"a": "*", "b": "file"}])
def test_find_files_wildcard(tmp_path, test_paths, find_kwargs):

    path_pattern = tmp_path / "{a}/foo"
    file_pattern = "{b}"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {
        "path": {
            0: str(tmp_path / "a1/foo/file"),
            1: str(tmp_path / "a2/foo/file"),
        },
        "a": {0: "a1", 1: "a2"},
        "b": {0: "file", 1: "file"},
    }
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_files(**find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files(find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files({"b": "XXX"}, **find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)


@pytest.mark.parametrize(
    "find_kwargs",
    [{"a": ["a1", "a2"], "b": "file"}, {"a": ["a1", "a2"], "b": ["file", "bar"]}],
)
def test_find_files_several(tmp_path, test_paths, find_kwargs):

    path_pattern = tmp_path / "{a}/foo"
    file_pattern = "{b}"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {
        "path": {
            0: str(tmp_path / "a1/foo/file"),
            1: str(tmp_path / "a2/foo/file"),
        },
        "a": {0: "a1", 1: "a2"},
        "b": {0: "file", 1: "file"},
    }
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_files(**find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files(find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files({"a": "XXX", "b": "XXX"}, **find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)


@pytest.mark.parametrize(
    "find_kwargs",
    [{"a": "a1"}, {"a": "a1", "b": "file"}, {"a": "a1", "b": ["file", "bar"]}],
)
def test_find_files_one_of_several(tmp_path, test_paths, find_kwargs):

    path_pattern = tmp_path / "{a}/foo"
    file_pattern = "{b}"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {
        "path": {0: str(tmp_path / "a1/foo/file")},
        "a": {0: "a1"},
        "b": {0: "file"},
    }
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_files(**find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files(find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files({"a": "XXX"}, **find_kwargs)
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_single_file(tmp_path, test_paths):

    path_pattern = tmp_path / "{a}/foo"
    file_pattern = "file"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    # error if more than one is found
    with pytest.raises(ValueError, match=r"Found more than one \(2\) files/ paths"):
        ff.find_single_file()

    # error if more than one is found
    with pytest.raises(ValueError, match="Found no files matching criteria"):
        ff.find_single_file(a="a3")

    expected = {"path": {0: str(tmp_path / "a1/foo/file")}, "a": {0: "a1"}}
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_single_file(a="a1")
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_single_file({"a": "a1"})
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_paths_scalar_number():

    ff = FileFinder(
        path_pattern="{path}", file_pattern="{file}", test_paths=["1/1", "2/2"]
    )

    index = pd.Index(["1/*"], name="path")
    expected = pd.DataFrame(["1"], columns=["path"], index=index)
    result = ff.find_paths(path=1)
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_files_scalar_number():

    ff = FileFinder(
        path_pattern="{path}", file_pattern="{file}", test_paths=["1/1", "2/2"]
    )

    index = pd.Index(["1/1"], name="path")
    expected = pd.DataFrame([["1", "1"]], columns=["path", "file"], index=index)
    result = ff.find_files(file=1)
    pd.testing.assert_frame_equal(result.df, expected)


def test_find_unparsable():

    ff = FileFinder("{cat}", "{cat}", test_paths=["a/b"])

    with pytest.raises(
        ValueError, match="Could not parse 'a/b' with the pattern '{cat}/{cat}'"
    ):
        ff.find_files()

    expected = pd.DataFrame([], columns=["cat"], index=pd.Index([], name="path"))

    with pytest.warns(match="Could not parse 'a/b' with the pattern '{cat}/{cat}'"):
        result = ff.find_files(on_parse_error="warn")
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files(on_parse_error="ignore")
    pd.testing.assert_frame_equal(result.df, expected)

    ff = FileFinder("{cat}", "{cat}", test_paths=["a/b", "a/a"])
    expected = {"path": {0: "a/a"}, "cat": {0: "a"}}
    expected = pd.DataFrame.from_dict(expected).set_index("path")
    result = ff.find_files(on_parse_error="ignore")
    pd.testing.assert_frame_equal(result.df, expected)

    with pytest.warns(match="Could not parse 'a/b' with the pattern '{cat}/{cat}'"):
        result = ff.find_files(on_parse_error="warn")
    pd.testing.assert_frame_equal(result.df, expected)

    with pytest.raises(
        ValueError, match="Could not parse 'a/b' with the pattern '{cat}/{cat}'"
    ):
        ff.find_files(on_parse_error="raise")

    with pytest.raises(
        ValueError,
        match="Unknown value for 'on_parse_error': 'foo'. Must be one of 'raise', 'warn' or 'ignore'.",
    ):
        ff.find_files(on_parse_error="foo")

    ff = FileFinder("", "{cat}_{cat}", test_paths=["a_b"])

    with pytest.raises(
        ValueError, match="Could not parse 'a_b' with the pattern '{cat}_{cat}'"
    ):
        ff.find_files()

    ff = FileFinder("{cat}_{cat}", "", test_paths=["a_b/"])

    with pytest.raises(
        ValueError, match="Could not parse 'a_b/' with the pattern '{cat}_{cat}'"
    ):
        ff.find_files()

    with pytest.raises(
        ValueError, match="Could not parse 'a_b/' with the pattern '{cat}_{cat}/'"
    ):
        ff.find_paths()



import textwrap

import pandas as pd
import pytest

from filefisher import FileFinder


@pytest.fixture(scope="module", params=["from_filesystem", "from_string"])
def test_paths(request, tmp_path):

    if request.param == "from_filesystem":
        return None

    paths = ["a1/foo/file", "a2/foo/file"]
    paths = [str(tmp_path / path) for path in paths]

    return paths


def test_pattern_no_fmt_spec():

    path_pattern = "{path:l}_{pattern:2d}_{no_fmt}/"
    file_pattern = "{file}_{pattern:2d}"

    path_pattern_no_fmt = "{path}_{pattern}_{no_fmt}/"
    file_pattern_no_fmt = "{file}_{pattern}"

    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    assert ff.path._pattern_no_fmt_spec == path_pattern_no_fmt
    assert ff.file._pattern_no_fmt_spec == file_pattern_no_fmt

    assert ff.full._pattern_no_fmt_spec == path_pattern_no_fmt + file_pattern_no_fmt


def test_keys():

    file_pattern = "{a:l}_{b}_{c:d}"
    path_pattern = "{ab}_{c:d}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    expected = ("ab", "c", "a", "b")
    assert ff.keys == expected

    expected = ("a", "b", "c")
    assert ff.keys_file == expected

    expected = ("ab", "c")
    assert ff.keys_path == expected


def test_repr():

    path_pattern = "/{a:l}/{b}"
    file_pattern = "{b}_{c:d}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    expected = """\
    <FileFinder>
    path_pattern: '/{a:l}/{b}/'
    file_pattern: '{b}_{c:d}'

    keys: 'a', 'b', 'c'
    """
    expected = textwrap.dedent(expected)
    assert expected == ff.__repr__()


def test_create_name():

    path_pattern = "{a:w}/{b}"
    file_pattern = "{b}_{c:l}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    result = ff.create_path_name(a="a", b="b")
    assert result == "a/b/"

    result = ff.create_file_name(b="b", c="c")
    assert result == "b_c"

    result = ff.create_full_name(a="a", b="b", c="c")
    assert result == "a/b/b_c"


def test_create_name_dict():

    path_pattern = "{a:w}/{b}"
    file_pattern = "{b}_{c:d}"
    ff = FileFinder(path_pattern=path_pattern, file_pattern=file_pattern)

    result = ff.create_path_name(dict(a="a", b="b"))
    assert result == "a/b/"

    result = ff.create_file_name(dict(b="b", c="c"))
    assert result == "b_c"

    result = ff.create_full_name(dict(a="a", b="b", c="c"))
    assert result == "a/b/b_c"


def test_find_paths_fmt():

    test_paths = ["a1/a1_abc", "ab200/ab200_aicdef"]

    path_pattern = "{letters:l}{num:d}"
    file_pattern = "{letters:l}{num:d}_{beg:2}{end}"

    ff = FileFinder(
        path_pattern=path_pattern, file_pattern=file_pattern, test_paths=test_paths
    )

    expected = {
        "path": {0: "a1/a1_abc", 1: "ab200/ab200_aicdef"},
        "letters": {0: "a", 1: "ab"},
        "num": {0: 1, 1: 200},
        "beg": {0: "ab", 1: "ai"},
        "end": {0: "c", 1: "cdef"},
    }
    expected = pd.DataFrame.from_dict(expected).set_index("path")

    result = ff.find_files()
    pd.testing.assert_frame_equal(result.df, expected)

    result = ff.find_files(num=[1])
    pd.testing.assert_frame_equal(result.df, expected.iloc[[0]])



import warnings

from filefisher import (
    __version__,  # noqa: F401
    _filefinder,
    _utils,
    cmip,
    filters,
)
from filefisher._filefinder import FileContainer, FileFinder

__all__ = [
    "_filefinder",
    "_utils",
    "cmip",
    "FileContainer",
    "FileFinder",
    "filters",
]

msg = (
    "`filefinder` has been renamed to `filefisher`! Please install filefisher to get "
    "future releases, and update your imports from `import filefinder` to "
    "`import filefisher`"
)

warnings.warn(msg, FutureWarning, stacklevel=2)



from importlib.metadata import version as _get_version

from filefisher import _filefinder, _utils, cmip, filters
from filefisher._filefinder import FileContainer, FileFinder

__all__ = [
    "_filefinder",
    "_utils",
    "cmip",
    "FileContainer",
    "FileFinder",
    "filters",
]

try:
    __version__ = _get_version("filefisher")
    del _get_version
except Exception:  # pragma: no cover
    # Local copy or not installed with setuptools.
    # Disable minimum version checks on downstream libraries.
    __version__ = "999"
